import math

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_empty,
    alifestd_mark_clade_duration_ratio_sister_asexual,
)


def test_empty():
    mt = alifestd_make_empty()
    mt["origin_time"] = None
    res = alifestd_mark_clade_duration_ratio_sister_asexual(mt)
    assert "clade_duration_ratio_sister" in res
    assert len(res) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "origin_time": [0, 10, 30],
        }
    )
    original_df = phylogeny_df.copy()
    with pytest.raises(ValueError):
        alifestd_mark_clade_duration_ratio_sister_asexual(
            phylogeny_df,
            mutate=mutate,
        )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[0]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    with pytest.raises(ValueError):
        alifestd_mark_clade_duration_ratio_sister_asexual(
            phylogeny_df,
            mutate=mutate,
        )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 4],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    with pytest.raises(ValueError):
        alifestd_mark_clade_duration_ratio_sister_asexual(
            phylogeny_df,
            mutate=mutate,
        )

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0]"],
            "origin_time": [0, 10, 30],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_duration_ratio_sister_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    assert result_df.loc[0, "clade_duration_ratio_sister"] == 1
    assert math.isnan(result_df.loc[1, "clade_duration_ratio_sister"])
    assert math.isnan(result_df.loc[2, "clade_duration_ratio_sister"])

    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple5(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3, 4, 5, 7],
            "ancestor_id": [0, 0, 0, 1, 1, 2, 2],
            "origin_time": [10, 0, 30, 20, 40, 50, 65],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mark_clade_duration_ratio_sister_asexual(
        phylogeny_df,
        mutate=mutate,
    )
    assert result_df.loc[0, "clade_duration_ratio_sister"] == 1
    assert result_df.loc[1, "clade_duration_ratio_sister"] == 30 / 35
    assert result_df.loc[2, "clade_duration_ratio_sister"] == 35 / 30
    assert math.isnan(result_df.loc[3, "clade_duration_ratio_sister"])
    assert math.isnan(result_df.loc[4, "clade_duration_ratio_sister"])
    assert math.isnan(result_df.loc[5, "clade_duration_ratio_sister"])
    assert math.isnan(result_df.loc[7, "clade_duration_ratio_sister"])

    if not mutate:
        assert original_df.equals(phylogeny_df)



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_mark_num_leaves_asexual import alifestd_mark_num_leaves_asexual
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_num_leaves_sibling_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Mark the number of leaves descendant from each node's siblings.

    Nodes with no siblings (e.g., root nodes) will have value 0 marked.

    Parameters
    ----------
    phylogeny_df : pd.DataFrame
        Alife standard DataFrame containing the phylogenetic relationships.

    mutate : bool, optional
        If True, modify the input DataFrame in place. Default is False.

    Returns
    -------
    pd.DataFrame
        Phylogeny DataFrame with an additional column "num_leaves_sibling"
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if "num_leaves" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_num_leaves_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["num_leaves_sibling"] = (
        phylogeny_df.loc[
            phylogeny_df["ancestor_id"],
            "num_leaves",
        ].values
        - phylogeny_df["num_leaves"].values
    )

    return phylogeny_df



import numpy as np

from ._jit import jit


@jit(nopython=True, nogil=True, cache=True)
def _fit_fblr(
    X_train: np.ndarray,
    y_train: np.ndarray,
    fit_intercept: bool,
    epsilon: float,
    lambda_ssr: float,
    f: float,
    gamma: float,
    convergenceTolerance: float,
    minimumIteration: int,
    maximumIteration: int,
    warnConvergence: bool,
) -> np.ndarray:
    """Implementation detail."""

    # adapted from
    # https://github.com/NurdanS/fblr/blob/241db04b0d200834407f363c12483dd553bcbf55/customClassifier.py

    if fit_intercept:
        X_train = np.hstack((np.ones((X_train.shape[0], 1)), X_train))

    n = X_train.shape[0]
    d = X_train.shape[1]

    I = np.identity(d)  # noqa: E741

    beta = 1e-8
    A = (X_train.T).dot(X_train) + beta * I
    b = (X_train.T).dot(y_train)
    w_0 = np.linalg.solve(A, b)

    doRegularization = lambda_ssr > 0 or gamma > 0

    log2 = np.log(2.0)
    q = 0.5

    v = (1 / n) * (X_train.T).dot(y_train - q)

    w = w_0.copy()
    for iteration in range(maximumIteration + 1):
        w_hat = w.copy()

        o = X_train.dot(w_hat)
        z = (np.log(1.0 + np.exp(o)) - log2 - 0.5 * o) / (o * o + epsilon)

        # weight update
        if doRegularization:
            p = np.ones(d)
            # if interceptColumnIndex >= 0 and interceptColumnIndex < d:
            #     p[interceptColumnIndex] = 0

            h = p / (np.abs(w_hat) ** (2 - f) + epsilon)
            H = np.diag(h.flatten())

            # A = (2/n) * (X_train.T Z X_train) + (lambda_ssr/d)*I + (gamma/d)*H
            A = (
                (2 / n) * np.multiply(z, X_train.T).dot(X_train)
                + (lambda_ssr / d) * I
                + (gamma / d) * H
            )
            b = (lambda_ssr / d) * w + v
            w = np.linalg.solve(A, b)
        else:
            # A = (2/n) * (X_train.T Z X_train)
            A = (2 / n) * np.multiply(z, X_train.T).dot(X_train)
            w = np.linalg.solve(A, v)

        change = np.max(np.abs(w - w_hat))
        if iteration >= minimumIteration and change <= convergenceTolerance:
            break
    else:
        if warnConvergence:
            print(
                "warning: maximum number of fblr iterations reached without "
                "convergence.",
            )

    return w


def fit_fblr(
    X_train: np.ndarray,
    y_train: np.ndarray,
    *,
    fit_intercept: bool = False,
    epsilon: float = 1e-10,
    lambda_ssr: float = 0,
    f: float = 0,
    gamma: float = 0.001,
    convergenceTolerance: float = 1e-3,
    minimumIteration: int = 2,
    maximumIteration: int = 10,
    warnConvergence: bool = True,
) -> np.ndarray:
    """Fit a logistic regression model using the Fast Binary Logistic
    Regression (FBLR) algorithm proposed by Saran and Nar (2025).

    Parameters
    ----------
    X_train : np.ndarray
        The input data, shape (n_samples, n_features).
    y_train : np.ndarray
        The target values, shape (n_samples,).
    fit_intercept : bool, default False
        Whether to include an intercept term in the model.
    epsilon : float, default 1e-10
        A small value to prevent division by zero.
    lambda_ssr : float, default 0
        Slow-step-regularization (SSR) parameter.
    f : float, default 0
        Regularization parameter for L_f norm penalty.
    gamma : float, default 0.001
        Regulariation parameter.
    convergenceTolerance : float, default 1e-3
        The tolerance for convergence.
    minimumIteration : int, default 2
        The minimum number of iterations to perform.
    maximumIteration : int, default 10
        The maximum number of iterations to perform.
    warnConvergence : bool, default True
        Whether to issue a warning if the maximum number of iterations is
        reached without convergence.

    Returns
    -------
    np.ndarray
        The estimated coefficients of the logistic regression model.

        If `fit_intercept` is True, the first element corresponds to the
        intercept term, and the remaining elements correspond to the
        coefficients for each feature in `X_train`. If `fit_intercept` is
        False, all elements correspond to the coefficients for each feature
        in `X_train`.

    Notes
    -----
    Anecdotally, approximations by this alborithm appear to systematically
    underestimate the true slope of the logistic regression line. However, they
    are still useful for estimating the sign and magnitude of the slope, and
    the algorithm is much faster than the standard logistic regression
    algorithm.

    References
    ----------
    Saran NA, Nar F. 2025. Fast binary logistic regression. PeerJ Computer
        Science 11:e2579 https://doi.org/10.7717/peerj-cs.2579
    """

    X_train = np.asarray(X_train, dtype=np.float64)
    y_train = np.asarray(y_train, dtype=np.float64)

    # handle simple singular case
    if np.ptp(X_train) < 1e-6:
        return np.zeros(X_train.shape[1])

    return _fit_fblr(
        X_train,
        y_train,
        fit_intercept=fit_intercept,
        epsilon=epsilon,
        lambda_ssr=lambda_ssr,
        f=f,
        gamma=gamma,
        convergenceTolerance=convergenceTolerance,
        minimumIteration=minimumIteration,
        maximumIteration=maximumIteration,
        warnConvergence=warnConvergence,
    )



import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact


@typing.overload  # type-hinting helper --- no runtime side-effects
def iter_mutual_ranks(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Iterator[int]:
    ...


@typing.overload  # type-hinting helper --- no runtime side-effects
def iter_mutual_ranks(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    compare: typing.Literal[False],
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Iterator[int]:
    ...


@typing.overload  # type-hinting helper --- no runtime side-effects
def iter_mutual_ranks(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    compare: typing.Literal[True],
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Iterator[typing.Tuple[int, bool]]:
    ...


@typing.overload  # type-hinting helper --- no runtime side-effects
def iter_mutual_ranks(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    compare: bool = False,
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Iterator[typing.Union[typing.Tuple[int, bool], int]]:
    ...


def iter_mutual_ranks(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    compare: bool = False,
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Iterator[typing.Union[typing.Tuple[int, bool], int]]:
    """Iterate over ranks with matching strata between columns in ascending
    order."""
    first_iter = zip(
        first.IterRetainedRanks(), first.IterRetainedDifferentia()
    )
    second_iter = zip(
        second.IterRetainedRanks(), second.IterRetainedDifferentia()
    )
    first_cur_rank, first_cur_differentia = next(first_iter)
    second_cur_rank, second_cur_differentia = next(second_iter)

    first_start_rank = first.GetRankAtColumnIndex(first_start_idx)
    second_start_rank = second.GetRankAtColumnIndex(second_start_idx)
    while first_cur_rank < first_start_rank:
        first_cur_rank, first_cur_differentia = next(first_iter)
    while second_cur_rank < second_start_rank:
        second_cur_rank, second_cur_differentia = next(second_iter)

    # a.k.a.
    # while (
    #     first_column_idx < first.GetNumStrataRetained()
    #     and second_column_idx < second.GetNumStrataRetained()
    # ):
    try:
        while True:
            if first_cur_rank == second_cur_rank:
                # strata at same rank can be compared
                if first_cur_differentia == second_cur_differentia:
                    # matching differentiae at the same rank,
                    # store rank and keep searching for mismatch
                    yield (first_cur_rank, True) if compare else first_cur_rank
                    # advance first
                    first_cur_rank, first_cur_differentia = next(first_iter)
                    # advance second
                    second_cur_rank, second_cur_differentia = next(second_iter)
                else:
                    # mismatching differentiae at the same rank
                    yield (
                        (
                            first_cur_rank,
                            False,
                        )
                        if compare
                        else first_cur_rank
                    )
            elif first_cur_rank < second_cur_rank:
                # current stratum on first column older than on second column
                # advance to next-newer stratum on first column
                first_cur_rank, first_cur_differentia = next(first_iter)
            elif first_cur_rank > second_cur_rank:
                # current stratum on second column older than on first column
                # advance to next-newer stratum on second column
                second_cur_rank, second_cur_differentia = next(second_iter)
    except StopIteration:
        return



import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_make_empty,
    alifestd_mask_descendants_asexual,
)


def test_empty():
    mt = alifestd_make_empty()
    result_df = alifestd_mask_descendants_asexual(
        mt, ancestor_mask=np.array([], dtype=bool)
    )
    assert "alifestd_mask_descendants_asexual" in result_df.columns
    assert len(result_df) == 0


@pytest.mark.parametrize("mutate", [True, False])
def test_simple1(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0]"],
            "origin_time": [0, 10, 30],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mask_descendants_asexual(
        phylogeny_df,
        mutate=mutate,
        ancestor_mask=np.array([False, True, False]),
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[0, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[1, "alifestd_mask_descendants_asexual"]
    assert not result_df.loc[2, "alifestd_mask_descendants_asexual"]
    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple2(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3],
            "ancestor_list": ["[0]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mask_descendants_asexual(
        phylogeny_df,
        mutate=mutate,
        ancestor_mask=np.array([True, False, True, False]),
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[0, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[1, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[2, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[3, "alifestd_mask_descendants_asexual"]
    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple3(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 4],
            "ancestor_list": ["[None]", "[None]", "[0]", "[1]"],
            "origin_time": [20, 10, 30, 45],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mask_descendants_asexual(
        phylogeny_df,
        mutate=mutate,
        ancestor_mask=np.array([True, False, True, False]),
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[0, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[1, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[2, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[4, "alifestd_mask_descendants_asexual"]
    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple4(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[0]"],
            "origin_time": [0, 10, 30],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mask_descendants_asexual(
        phylogeny_df,
        mutate=mutate,
        ancestor_mask=np.array([False, True, False]),
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[0, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[1, "alifestd_mask_descendants_asexual"]
    assert not result_df.loc[2, "alifestd_mask_descendants_asexual"]
    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple5(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0, 2, 3, 4],
            "ancestor_list": ["[0]", "[None]", "[0]", "[1]", "[1]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mask_descendants_asexual(
        phylogeny_df,
        mutate=mutate,
        ancestor_mask=np.array([True, False, False, True, False]),
    )
    result_df.index = result_df["id"]
    assert result_df.loc[1, "alifestd_mask_descendants_asexual"]
    assert not result_df.loc[0, "alifestd_mask_descendants_asexual"]
    assert not result_df.loc[2, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[3, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[4, "alifestd_mask_descendants_asexual"]
    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple6(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [1, 0],
            "ancestor_list": ["[None]", "[None]"],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mask_descendants_asexual(
        phylogeny_df,
        mutate=mutate,
        ancestor_mask=np.array([False, True]),
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[1, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[0, "alifestd_mask_descendants_asexual"]
    if not mutate:
        assert original_df.equals(phylogeny_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_simple7(mutate: bool):
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_list": ["[None]", "[0]", "[1]"],
            "origin_time": [0, 10, 30],
        }
    )
    original_df = phylogeny_df.copy()
    result_df = alifestd_mask_descendants_asexual(
        phylogeny_df,
        mutate=mutate,
        ancestor_mask=np.array([False, True, True]),
    )
    result_df.index = result_df["id"]
    assert not result_df.loc[0, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[1, "alifestd_mask_descendants_asexual"]
    assert result_df.loc[2, "alifestd_mask_descendants_asexual"]
    if not mutate:
        assert original_df.equals(phylogeny_df)



import numpy as np
import pytest

from hstrat._auxiliary_lib import pairwise
from hstrat.hstrat import recency_proportional_resolution_curbed_algo


@pytest.mark.parametrize(
    "size_curb",
    [
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "time_sequence",
    [
        range(10**3),
        (i for i in range(10**2) for __ in range(2)),
        np.random.default_rng(1).integers(
            low=0,
            high=2**32,
            size=10,
        ),
        (2**32,),
    ],
)
def test_ranks_sorted_and_unique(size_curb, time_sequence):
    policy = recency_proportional_resolution_curbed_algo.Policy(size_curb)
    spec = policy.GetSpec()
    instance = recency_proportional_resolution_curbed_algo.IterRetainedRanks(
        spec
    )
    for num_strata_deposited in time_sequence:
        for which in (
            instance,
            recency_proportional_resolution_curbed_algo.IterRetainedRanks(
                spec
            ),
        ):
            assert all(
                i < j
                for i, j in pairwise(
                    which(
                        policy,
                        num_strata_deposited,
                    )
                )
            )



import os
import pathlib
import subprocess

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def test_surface_build_tree_cli_help():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_build_tree",
            "--help",
        ],
        check=True,
    )


def test_surface_build_tree_cli_version():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_build_tree",
            "--version",
        ],
        check=True,
    )


def test_surface_build_tree_cli_csv():
    output_file = "/tmp/hstrat_surface_build_tree.csv"
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_build_tree",
            output_file,
        ],
        check=True,
        input=f"{assets}/packed.csv".encode(),
    )
    assert os.path.exists(output_file)


def test_surface_build_tree_cli_parquet():
    output_file = "/tmp/hstrat_surface_build_tree.pqt"
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat.dataframe.surface_build_tree",
            output_file,
            "--shrink-dtypes",
            "--collapse-unif-freq",
            "0",
            "--exploded-slice-size",
            "50_000_000",
        ],
        check=True,
        input=f"{assets}/packed.csv".encode(),
    )
    assert os.path.exists(output_file)



import numpy as np


# adapted from https://stackoverflow.com/a/30489294/17332200
def fill_zeros_with_last(arr: np.ndarray) -> np.ndarray:
    """Replace zeros in array with the nearest preceding non-zero value."""
    prev = np.arange(len(arr))
    prev[arr == 0] = 0
    prev = np.maximum.accumulate(prev)
    return arr[prev]



import gc
import logging
import typing

import pandas as pd
import polars as pl
from tqdm import tqdm

from .._auxiliary_lib import (
    alifestd_assign_contiguous_ids,
    alifestd_collapse_unifurcations,
    alifestd_delete_trunk_asexual,
    alifestd_prefix_roots,
    get_sole_scalar_value_polars,
    log_context_duration,
    log_memory_usage,
    render_pandas_snapshot,
    render_polars_snapshot,
)
from ..phylogenetic_inference.tree.trie_postprocess import NopTriePostprocessor


def _do_collapse_unifurcations(
    df: pd.DataFrame,
) -> pd.DataFrame:
    with log_context_duration("alifestd_assign_contiguous_ids", logging.info):
        df = alifestd_assign_contiguous_ids(df, mutate=True)

    gc.collect()

    with log_context_duration("alifestd_collapse_unifurcations", logging.info):
        df = alifestd_collapse_unifurcations(df, mutate=True)

    render_pandas_snapshot(df, "collapsed tree", logging.info)

    gc.collect()

    return df


def _do_delete_trunk(
    df: pd.DataFrame,
) -> pd.DataFrame:
    with log_context_duration("alifestd_assign_contiguous_ids", logging.info):
        df = alifestd_assign_contiguous_ids(df, mutate=True)

    gc.collect()

    with log_context_duration(
        "df['dstream_S'].unique().squeeze()", logging.info
    ):
        dstream_S = df["dstream_S"].unique().squeeze()

    df["is_trunk"] = df["hstrat_rank"] < df["dstream_S"]
    df["origin_time"] = df["hstrat_rank"]

    with log_context_duration("alifestd_delete_trunk_asexual", logging.info):
        df = alifestd_delete_trunk_asexual(df, mutate=True)

    with log_context_duration("alifestd_assign_contiguous_ids", logging.info):
        df = alifestd_assign_contiguous_ids(df, mutate=True)

    with log_context_duration("alifestd_prefix_roots", logging.info):
        # extend newly-clipped roots all the way back to dstream_S boundary
        df = alifestd_prefix_roots(
            df, allow_id_reassign=True, origin_time=dstream_S, mutate=True
        )

    del df["is_trunk"]
    del df["ancestor_is_trunk"]
    del df["origin_time"]
    gc.collect()

    return df


def _do_assign_contiguous_ids(
    df: pd.DataFrame,
) -> pd.DataFrame:
    with log_context_duration("alifestd_assign_contiguous_ids", logging.info):
        df = alifestd_assign_contiguous_ids(df, mutate=True)

    gc.collect()

    render_pandas_snapshot(df, "reassigned tree", logging.info)

    return df


def _surface_postprocess_trie_via_pandas(
    df: pl.DataFrame,
    *,
    trie_postprocessor: typing.Callable = NopTriePostprocessor(),
    delete_trunk: bool = True,
    # ^^^ NopTriePostprocessor is stateless, so is safe as default value
) -> pl.DataFrame:
    """Legacy implementation of trie postprocessing using Pandas, kept for
    testing purposes. Should not be used in production code."""
    logging.info("beginning surface_postprocess_trie")
    log_memory_usage(logging.info)
    render_polars_snapshot(df, "raw tree", logging.info)

    logging.info("extracting differentia bitwidth")
    differentia_bitwidth = get_sole_scalar_value_polars(
        df, "hstrat_differentia_bitwidth"
    )
    logging.info(f" - differentia bitwidth: {differentia_bitwidth}")

    logging.info("converting DataFrame to Pandas...")
    with log_context_duration("pl.DataFrame.to_pandas", logging.info):
        df = df.lazy().collect().to_pandas()
    render_pandas_snapshot(df, "as pandas", logging.info)
    log_memory_usage(logging.info)
    original_columns = df.columns

    if delete_trunk:
        df = _do_delete_trunk(df)

    df = _do_collapse_unifurcations(df)

    df = _do_assign_contiguous_ids(df)

    with log_context_duration("trie_postprocessor", logging.info):
        pre_postprocessor_columns = {*df.columns}
        df = df.rename(columns={"hstrat_rank": "rank"})
        df = trie_postprocessor(
            df,
            p_differentia_collision=2**-differentia_bitwidth,
            mutate=True,
            progress_wrap=tqdm,
        )
        df = df.rename(columns={"rank": "hstrat_rank"})

    render_pandas_snapshot(df, "with trie postprocessing", logging.info)

    logging.info("setting up hstrat_rank_from_t0...")
    df["hstrat_rank_from_t0"] = df["hstrat_rank"] - df["dstream_S"]

    to_keep = {*original_columns} - {
        "dstream_S",
        "hstrat_differentia_bitwidth",
        "hstrat_rank",
    }
    to_drop = pre_postprocessor_columns - to_keep
    logging.info(f"dropping columns {to_drop=}...")
    df.drop(columns=[*to_drop], inplace=True)
    gc.collect()

    logging.info("converting DataFrame to Polars...")
    with log_context_duration("pl.from_pandas", logging.info):
        df = pl.from_pandas(df)
    gc.collect()
    render_polars_snapshot(df, "as polars", logging.info)
    log_memory_usage(logging.info)

    return df



import typing

from interval_search import binary_search

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...frozen_instrumentation import HereditaryStratigraphicSpecimen
from ...genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratigraphicSurface,
)
from .._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._calc_rank_of_first_retained_disparity_between_generic import (
    calc_rank_of_first_retained_disparity_between_generic,
)


def calc_rank_of_first_retained_disparity_between_bsearch(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find first mismatching strata between columns.

    Implementation detail. Provides optimized implementation for special
    case where both self and second use the perfect resolution stratum
    retention policy.
    """
    if isinstance(first, HereditaryStratigraphicSpecimen) or isinstance(
        second, HereditaryStratigraphicSpecimen
    ):
        raise NotImplementedError(
            "HereditaryStratigraphicSpecimen not yet supported",
        )

    # both must have (effectively) used the perfect resolution policy
    assert not first.HasDiscardedStrata() and not second.HasDiscardedStrata()

    lower_bound = 0
    upper_bound = min(
        [
            first.GetNumStrataDeposited() - 1,
            second.GetNumStrataDeposited() - 1,
        ]
    )
    assert lower_bound <= upper_bound

    def differentia_at(
        which: typing.Union[
            HereditaryStratigraphicColumn, HereditaryStratigraphicSurface
        ],
        idx: int,
    ) -> int:
        return which.GetStratumAtColumnIndex(idx).GetDifferentia()

    def predicate(idx: int) -> bool:
        return differentia_at(first, idx) != differentia_at(second, idx)

    first_disparite_idx = binary_search(
        predicate,
        lower_bound,
        upper_bound,
    )

    if first_disparite_idx is not None:
        # disparate strata found
        assert (
            first.GetStratumDifferentiaBitWidth()
            == second.GetStratumDifferentiaBitWidth()
        )
        collision_plausibility_threshold = (
            calc_min_implausible_spurious_consecutive_differentia_collisions_between(
                first,
                second,
                significance_level=1.0 - confidence_level,
            )
            - 1
        )
        # discount collision_implausibility_threshold - 1 common
        # ranks due to potential spurious differentia collisions;
        # if not enough common ranks are available we still know
        # *definitively* that a disparity occured (because we
        # observed disparite strata at the same rank); so, make the
        # conservative assumption that the disparity occured as far
        # back as possible (rank 0)
        spurious_collision_corrected_idx = max(
            first_disparite_idx - collision_plausibility_threshold,
            0,
        )
        first_disparite_rank = first.GetRankAtColumnIndex(
            spurious_collision_corrected_idx
        )
        assert first_disparite_rank >= 0
        return first_disparite_rank
    else:
        # no disparate strata found
        # fall back to calc_rank_of_first_retained_disparity_between_generic to
        # handle proper bookkeeping in this case while skipping most of the
        # search
        return calc_rank_of_first_retained_disparity_between_generic(
            first,
            second,
            first_start_idx=upper_bound,
            second_start_idx=upper_bound,
            confidence_level=confidence_level,
        )



#!/usr/bin/env python3

import argparse
import itertools
import os
import subprocess
import sys
import typing

from Bio.Phylo.BaseTree import Clade as BioClade
import alifedata_phyloinformatics_convert as apc
from colorclade import draw_colorclade_tree
import matplotlib.pyplot as plt
import opytional as opyt
import pandas as pd
from teeplot import teeplot as tp

from hstrat._auxiliary_lib import (
    alifestd_calc_triplet_distance_asexual,
    alifestd_collapse_unifurcations,
    alifestd_count_leaf_nodes,
    alifestd_mark_node_depth_asexual,
    alifestd_prune_extinct_lineages_asexual,
    alifestd_try_add_ancestor_list_col,
)


def to_ascii(
    phylogeny_df: pd.DataFrame, taxon_labels: typing.List[str]
) -> str:
    """Print a sample of the phylogeny dataframe."""
    phylogeny_df = phylogeny_df.copy()
    phylogeny_df["extant"] = phylogeny_df["taxon_label"].isin(taxon_labels)
    phylogeny_df = alifestd_prune_extinct_lineages_asexual(
        phylogeny_df, mutate=True
    ).drop(columns=["extant"])
    phylogeny_df = alifestd_collapse_unifurcations(phylogeny_df, mutate=True)

    dp_tree = apc.RosettaTree(phylogeny_df).as_dendropy
    for nd in dp_tree.preorder_node_iter():
        nd._child_nodes.sort(
            key=lambda nd: max(leaf.taxon.label for leaf in nd.leaf_iter()),
        )
    return dp_tree.as_ascii_plot()


def load_df(path: str) -> pd.DataFrame:
    """Helper function to get the appropriate pandas dataframe save handler."""
    df_load_handlers = {
        ".csv": pd.read_csv,
        ".json": pd.read_json,
        ".pqt": pd.read_parquet,
        ".parquet": pd.read_parquet,
    }
    ext = os.path.splitext(path)[-1]
    try:
        return df_load_handlers[ext](path)
    except KeyError:
        raise ValueError(f"Unsupported dataframe path extension: {ext}")


def sample_reference_and_reconstruction(
    differentia_bitwidth: int,
    surface_size: int,
    fossil_interval: typing.Optional[int],
) -> typing.Dict[str, pd.DataFrame]:
    """Sample a reference phylogeny and corresponding reconstruction."""
    try:
        paths = subprocess.run(
            [
                f"{os.path.dirname(__file__)}/"
                "end2end_tree_reconstruction_with_dstream_surf.sh",
                "--differentia-bitwidth",
                f"{differentia_bitwidth}",
                "--surface-size",
                f"{surface_size}",
                *(
                    ["--fossil-interval", f"{fossil_interval}"]
                    * (fossil_interval is not None)
                ),
            ],
            check=True,
            capture_output=True,
            text=True,
        ).stdout.strip()
    except subprocess.CalledProcessError as e:
        print(f"\033[33m{e.stdout}\033[0m")  # color yellow
        print(f"\033[31m{e.stderr}\033[0m")  # color red
        raise e

    path_vars = dict()  # outparam for exec
    exec(paths, path_vars)  # hack to load paths from shell script output
    true_phylo_df = load_df(path_vars["true_phylo_df_path"])
    reconst_phylo_df = alifestd_try_add_ancestor_list_col(
        load_df(path_vars["reconst_phylo_df_path"]),
    )  # ancestor_list column must be added to comply with alife standard

    assert alifestd_count_leaf_nodes(
        true_phylo_df
    ) == alifestd_count_leaf_nodes(reconst_phylo_df)

    # we use == False and == True here because there are NaN values as well
    reconst_phylo_df_extant = reconst_phylo_df.copy()
    reconst_phylo_df_extant["extant"] = (
        reconst_phylo_df["is_fossil"] == False  # noqa: E712
    )
    reconst_phylo_df_no_fossils = alifestd_prune_extinct_lineages_asexual(
        reconst_phylo_df_extant,
    )

    taxa_to_drop = reconst_phylo_df["taxon_label"][
        reconst_phylo_df["is_fossil"] == True  # noqa: E712
    ]
    new_df = (
        true_phylo_df.set_index("taxon_label")
        .drop(taxa_to_drop)  # type: ignore
        .reset_index()
    )

    true_phylo_df_no_fossils = alifestd_prune_extinct_lineages_asexual(new_df)

    return {
        "exact": true_phylo_df,
        "reconst": reconst_phylo_df,
        "exact_dropped_fossils": true_phylo_df_no_fossils,
        "reconst_dropped_fossils": reconst_phylo_df_no_fossils,
    }


def _label_func(node: BioClade):
    return " " + node.name if node.name and hash(node.name) % 12 == 0 else ""


def plot_colorclade_comparison(frames: typing.Dict[str, pd.DataFrame]) -> None:
    plotter_kwargs = {
        "taxon_name_key": "taxon_label",
        "backend": "biopython",
        "color_labels": "white",
        "line_width": 2.5,
    }
    plt.style.use("dark_background")
    fig, axes = plt.subplots(3, 2)

    frames["exact"] = alifestd_collapse_unifurcations(frames["exact"])
    frames["exact"]["origin_time"] = frames["exact"]["depth"]
    frames["reconst"]["origin_time"] = frames["reconst"]["hstrat_rank"]

    frames["exact_dropped_fossils"] = alifestd_collapse_unifurcations(
        frames["exact_dropped_fossils"],
    )
    frames["exact_dropped_fossils"]["origin_time"] = frames[
        "exact_dropped_fossils"
    ]["depth"]
    frames["reconst_dropped_fossils"]["origin_time"] = frames[
        "reconst_dropped_fossils"
    ]["hstrat_rank"]

    true_df_no_lengths = alifestd_mark_node_depth_asexual(frames["exact"])
    reconst_df_no_lengths = alifestd_mark_node_depth_asexual(frames["reconst"])
    true_df_no_lengths["origin_time"] = true_df_no_lengths["node_depth"]
    reconst_df_no_lengths["origin_time"] = reconst_df_no_lengths["node_depth"]

    for i, frame in enumerate(
        (
            frames["exact_dropped_fossils"],
            frames["reconst_dropped_fossils"],
            frames["exact"],
            frames["reconst"],
            true_df_no_lengths,
            reconst_df_no_lengths,
        )
    ):
        draw_colorclade_tree(
            frame,
            ax=axes.flat[i],
            **plotter_kwargs,
            label_tips=_label_func if i % 2 == 0 else False,
        )
        axes.flat[i].set_xlim(0, max(frame["origin_time"].unique()))

    # apply a polynomial scale to first two plots, which effectively cause
    # more of the graph to focus on recent data
    receny_scale_funcs = (lambda x: x**10, lambda x: x**0.1)
    axes.flat[0].set_xscale("function", functions=receny_scale_funcs)
    axes.flat[1].set_xscale("function", functions=receny_scale_funcs)

    # flip orientation of right-hand panels
    for i in range(1, len(axes.flat), 2):
        axes.flat[i].set_xlim(reversed(axes.flat[i].get_xlim()))

    fig.set_size_inches(10 * axes.shape[1], 10 * axes.shape[0])
    for i, title in enumerate(
        (
            "True Phylogeny Dropped Fossils Recency Scaled",
            "Reconstructed Phylogeny Dropped Fossils Recency Scaled",
            "True Phylogeny Time Scaled",
            "Reconstructed Phylogeny Time Scaled",
            "True Phylogeny Topology Only",
            "Reconstructed Phylogeny Topology Only",
        )
    ):
        axes.flat[i].set_title(title)
    plt.tight_layout()


def display_reconstruction(
    frames: typing.Dict[str, pd.DataFrame],
    *,
    create_plots: bool,
    **kwargs,
) -> None:
    """Print a sample of the reference and reconstructed phylogenies."""
    show_taxa = (
        frames["reconst_dropped_fossils"]["taxon_label"]
        .dropna()
        .sample(6, random_state=1)
    )
    print("ground-truth phylogeny sample:")
    print(to_ascii(frames["exact_dropped_fossils"], show_taxa))
    print()
    print("reconstructed phylogeny sample:")
    print(to_ascii(frames["reconst_dropped_fossils"], show_taxa))

    if create_plots:
        for df in frames.values():
            df["taxon_label"] = df["taxon_label"].apply(lambda x: x and x[:6])
        tp.tee(
            plot_colorclade_comparison,
            {
                k: alifestd_try_add_ancestor_list_col(v)
                for k, v in frames.items()
            },
            teeplot_dpi=100,
            teeplot_outattrs=kwargs,
            teeplot_outdir="/tmp/end2end-visualizations",
        )


def test_reconstruct_one(
    differentia_bitwidth: int,
    surface_size: int,
    fossil_interval: typing.Optional[int],
    *,
    visualize: bool,
) -> typing.Dict[str, typing.Union[int, float, None]]:
    """Test the reconstruction of a single phylogeny."""
    print("=" * 80)
    print(f"surface_size: {surface_size}")
    print(f"differentia_bitwidth: {differentia_bitwidth}")
    print(f"fossil_interval: {fossil_interval}")

    frames = sample_reference_and_reconstruction(
        differentia_bitwidth,
        surface_size,
        fossil_interval,
    )

    display_reconstruction(
        frames,
        differentia_bitwidth=differentia_bitwidth,
        surface_size=surface_size,
        fossil_interval=fossil_interval,
        create_plots=visualize,
    )
    reconstruction_error = alifestd_calc_triplet_distance_asexual(
        alifestd_collapse_unifurcations(frames["exact"]), frames["reconst"]
    )

    reconstruction_error_dropped_fossils = (
        alifestd_calc_triplet_distance_asexual(
            alifestd_collapse_unifurcations(frames["exact_dropped_fossils"]),
            frames["reconst_dropped_fossils"],
        )
    )

    print(f"{reconstruction_error=}")
    print(f"{reconstruction_error_dropped_fossils=}")
    assert 0 <= reconstruction_error <= 1  # should be in the range [0,1]

    return {
        "differentia_bitwidth": differentia_bitwidth,
        "surface_size": surface_size,
        "fossil_interval": fossil_interval,
        "error": reconstruction_error,
        "error_dropped_fossils": reconstruction_error_dropped_fossils,
    }


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--skip-visualization", action="store_true")
    return parser.parse_args()


if __name__ == "__main__":
    sys.setrecursionlimit(100000)
    args = _parse_args()
    reconstruction_error_results = pd.DataFrame(
        [
            test_reconstruct_one(
                differentia_bitwidth,
                surface_size,
                fossil_interval,
                visualize=not args.skip_visualization,
            )
            for (
                fossil_interval,
                surface_size,
                differentia_bitwidth,
            ) in itertools.product((None, 50, 200), (256, 64, 16), (64, 8, 1))
        ]
    ).sort_values(
        ["fossil_interval", "surface_size", "differentia_bitwidth"],
        ascending=False,
    )
    reconstruction_error_results.to_csv(
        "/tmp/end2end-reconstruction-error.csv",
    )

    # error should increase with decreasing surface size
    tolerance = 0.02
    for f, x in reconstruction_error_results.groupby("fossil_interval"):
        for first, second in itertools.pairwise(x.itertuples()):
            if second.error_dropped_fossils < first.error_dropped_fossils:  # type: ignore
                msg = (
                    f"Reconstruction error of {first.error_dropped_fossils} from run "  # type: ignore
                    f"{first.differentia_bitwidth}-{first.surface_size}-{opyt.apply_if(first.fossil_interval, int)} "  # type: ignore
                    f" unexpectedly higher than {second.error_dropped_fossils} from run "  # type: ignore
                    f"{second.differentia_bitwidth}-{second.surface_size}-{opyt.apply_if(second.fossil_interval, int)}"  # type: ignore
                )
                if (
                    first.error_dropped_fossils - second.error_dropped_fossils  # type: ignore
                    < tolerance
                ):
                    print(msg)
                    print(
                        "Difference is within error tolerance, continuing..."
                    )
                else:
                    raise ValueError(msg)



import polars as pl

from ._alifestd_assign_contiguous_ids import _reassign_ids_asexual


def alifestd_assign_contiguous_ids_polars(
    phylogeny_df: pl.DataFrame,
) -> pl.DataFrame:
    """Reassign so each organism's id corresponds to its row number.

    Organisms retain the same row location; only id numbers change. Input
    dataframe is not mutated by this operation unless `mutate` True.
    """
    if "ancestor_list" in phylogeny_df.columns:
        raise NotImplementedError

    phylogeny_df = phylogeny_df.lazy().collect()

    new_ancestor_ids = _reassign_ids_asexual(
        phylogeny_df["id"].to_numpy(),
        phylogeny_df["ancestor_id"].to_numpy(),
    )

    return (
        phylogeny_df.drop("id")
        .with_row_index("id")
        .with_columns(
            ancestor_id=pl.Series(new_ancestor_ids),
        )
    )



import numpy as np
import pandas as pd

from ._alifestd_calc_clade_trait_count_asexual import (
    alifestd_calc_clade_trait_count_asexual,
)


def alifestd_calc_clade_trait_frequency_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    mask_trait_absent: np.ndarray,  # 1D boolean array
    mask_trait_present: np.ndarray,  # 1D boolean array
) -> np.ndarray:
    """Calculate what fraction of nodes within each clade have a given trait.

    Clades are defined as a node and all descendant nodes. The
    `mask_trait_absent` parameter can be used to exclude nodes from
    consideration, for instance &'ing with `alifestd_mark_leaves` can be used
    to only consider trait frequency among descendant leaves.

    Returns a numpy array of the same length as the input DataFrame, with array
    elements as the number of nodes in the clade that have the trait. Returned
    array matches row order of the input DataFrame.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    mask_trait_absent = np.asarray(mask_trait_absent, dtype=bool)
    mask_trait_present = np.asarray(mask_trait_present, dtype=bool)

    if np.any(mask_trait_absent & mask_trait_present):
        raise ValueError(
            "mask_trait_absent and mask_trait_present not mutually exclusive",
        )

    trait_count_absent = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=True,
        trait_mask=mask_trait_absent,
    )
    trait_count_present = alifestd_calc_clade_trait_count_asexual(
        phylogeny_df,
        mutate=True,
        trait_mask=mask_trait_present,
    )

    return trait_count_present / (
        trait_count_present + trait_count_absent
    )  # numpy division implicitly converts to float



import numpy as np
import pandas as pd

from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_mark_leaves import alifestd_mark_leaves


def alifestd_add_inner_leaves(
    phylogeny_df: pd.DataFrame, mutate: bool = False
) -> pd.DataFrame:
    """Create a zero-length branch with leaf node for each inner node.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """
    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_mark_leaves(phylogeny_df, mutate=True)

    inner_df = phylogeny_df[~phylogeny_df["is_leaf"]].copy()
    inner_df["is_leaf"] = True

    if inner_df.empty:
        return phylogeny_df

    if "is_root" in inner_df:
        inner_df["is_root"] = False

    if "origin_time_delta" in inner_df:
        inner_df["origin_time_delta"] = 0

    inner_df["ancestor_id"] = inner_df["id"]

    inner_df["id"] = np.arange(len(inner_df)) + phylogeny_df["id"].max() + 1
    if not (inner_df["id"].min() > phylogeny_df["id"].max()):
        print(inner_df["id"].min(), phylogeny_df["id"].max())
        raise ValueError("overflow in new id assigment")

    if "ancestor_list" in inner_df:
        inner_df["ancestor_list"] = alifestd_make_ancestor_list_col(
            inner_df["id"],
            inner_df["ancestor_id"],
        )

    if "ancestor_id" not in phylogeny_df:
        del inner_df["ancestor_id"]

    return pd.concat([phylogeny_df, inner_df], ignore_index=True)



import pandas as pd
import polars as pl

from hstrat._auxiliary_lib import coerce_to_pandas


def test_coerce_to_pandas_polars_lazyframe():
    lf = pl.LazyFrame({"a": [1, 2, 3]})
    result = coerce_to_pandas(lf)

    assert isinstance(result, pd.DataFrame)
    pd.testing.assert_frame_equal(result, pd.DataFrame({"a": [1, 2, 3]}))


def test_coerce_to_pandas_polars_dataframe():
    df = pl.DataFrame({"x": [10, 20], "y": [30, 40]})
    result = coerce_to_pandas(df)
    assert isinstance(result, pd.DataFrame)
    pd.testing.assert_frame_equal(
        result, pd.DataFrame({"x": [10, 20], "y": [30, 40]})
    )


def test_coerce_to_pandas_recurse_iterable():
    data = (pl.DataFrame({"a": [1, 2]}), 42, pl.DataFrame({"b": [3, 4]}))
    result = coerce_to_pandas(data, recurse=True)
    assert isinstance(result, tuple)
    assert len(result) == 3
    assert isinstance(result[0], pd.DataFrame)
    assert result[1] == 42
    assert isinstance(result[2], pd.DataFrame)


def test_coerce_to_pandas_no_coercion_needed():
    data = [1, 2, 3, "no-polars-here"]
    result = coerce_to_pandas(data)
    assert result == data



import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_strictly_bifurcating_asexual import (
    alifestd_is_strictly_bifurcating_asexual,
)
from ._alifestd_mark_is_left_child_asexual import (
    alifestd_mark_is_left_child_asexual,
)
from ._alifestd_mark_is_right_child_asexual import (
    alifestd_mark_is_right_child_asexual,
)
from ._alifestd_mark_left_child_asexual import alifestd_mark_left_child_asexual
from ._alifestd_mark_right_child_asexual import (
    alifestd_mark_right_child_asexual,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_mark_sister_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `sister`, containing the id of each node's sibling.

    Root nodes will be marked with their own id. Phylogeny must be
    strictly bifurcating.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if not alifestd_is_strictly_bifurcating_asexual(phylogeny_df):
        raise ValueError("phylogeny_df must be strictly bifurcating")

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if "left_child_id" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_left_child_asexual(
            phylogeny_df, mutate=True
        )
    if "right_child_id" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_right_child_asexual(
            phylogeny_df, mutate=True
        )
    if "is_left_child" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_is_left_child_asexual(
            phylogeny_df, mutate=True
        )
    if "is_right_child" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_is_right_child_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["sister_id"] = (
        phylogeny_df.loc[phylogeny_df["ancestor_id"], "right_child_id"].values
        * phylogeny_df["is_left_child"].values
        + phylogeny_df.loc[phylogeny_df["ancestor_id"], "left_child_id"].values
        * phylogeny_df["is_right_child"].values
        + phylogeny_df["id"].values
        * (phylogeny_df["ancestor_id"].values == phylogeny_df["id"].values)
    )

    return phylogeny_df



import numpy as np
import pytest
from scipy.special import expit

from hstrat._auxiliary_lib._fit_fblr import fit_fblr


@pytest.mark.parametrize(
    "n_samples, min_corr",
    [
        (10, 0.4),  # very noisy -> weak positive correlation
        (50, 0.6),  # small -> moderate correlation
        (500, 0.8),  # medium -> strong correlation
        (5000, 0.95),  # large -> very strong correlation
    ],
)
def test_beta_estimate_correlation_across_slopes(
    n_samples: int, min_corr: float
):
    """
    For each sample size, generate many different true slopes beta,
    simulate y ~ Bernoulli(expit(X * beta)), fit logistic regression
    (no intercept), and check that the Pearson correlation between
    [beta_true] and [beta_estimated] exceeds min_corr.
    """
    rng = np.random.RandomState(0)

    # choose a grid of true betas to span weak to strong effects
    true_betas = np.linspace(-3.0, 5.0, 12)

    est_betas = []
    for beta in true_betas:
        # simulate
        X = rng.randn(n_samples, 1)
        logits = X.flatten() * beta
        y = rng.binomial(1, expit(logits))

        # fit
        w = fit_fblr(
            X_train=X,
            y_train=y,
            fit_intercept=False,
        )
        assert w.shape == (1,)
        est_betas.append(w[0])

    corr_matrix = np.corrcoef(true_betas, est_betas)
    corr = corr_matrix[0, 1]

    assert corr >= min_corr



import pickle
import tempfile

import pytest

from hstrat.hstrat import pseudostochastic_algo


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_init(hash_salt):
    assert (
        pseudostochastic_algo.Policy(hash_salt).GetSpec()
        == pseudostochastic_algo.Policy(
            policy_spec=pseudostochastic_algo.PolicySpec(hash_salt),
        ).GetSpec()
    )

    policy = pseudostochastic_algo.Policy(hash_salt)

    # invariants
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBound)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBound)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundPessimalRank)
    assert callable(policy.CalcNumStrataRetainedUpperBound)
    # scrying
    assert policy.CalcMrcaUncertaintyAbsExact is None
    assert policy.CalcMrcaUncertaintyRelExact is None
    assert policy.CalcNumStrataRetainedExact is None
    assert policy.CalcRankAtColumnIndex is None
    assert policy.IterRetainedRanks is None
    # enactment
    assert callable(policy.GenDropRanks)


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_eq(hash_salt):
    policy = pseudostochastic_algo.Policy(hash_salt)
    assert policy == policy
    assert policy == pseudostochastic_algo.Policy(hash_salt)
    assert policy == policy.WithoutCalcRankAtColumnIndex()
    assert (
        policy.WithoutCalcRankAtColumnIndex()
        == policy.WithoutCalcRankAtColumnIndex()
    )
    assert not policy == pseudostochastic_algo.Policy(hash_salt + 1)


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_pickle(hash_salt):
    original = pseudostochastic_algo.Policy(hash_salt)
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_GetSpec(hash_salt):
    assert pseudostochastic_algo.Policy(
        hash_salt
    ).GetSpec() == pseudostochastic_algo.PolicySpec(hash_salt)


@pytest.mark.parametrize(
    "hash_salt",
    [
        1,
        2,
        3,
        7,
        42,
        100,
    ],
)
def test_WithoutCalcRankAtColumnIndex(hash_salt):

    original = pseudostochastic_algo.Policy(hash_salt)
    stripped = original.WithoutCalcRankAtColumnIndex()

    assert stripped.CalcRankAtColumnIndex is None

    assert (
        original.CalcMrcaUncertaintyAbsUpperBound
        == stripped.CalcMrcaUncertaintyAbsUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBound
        == stripped.CalcMrcaUncertaintyRelUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundPessimalRank
    )
    assert (
        original.CalcNumStrataRetainedUpperBound
        == stripped.CalcNumStrataRetainedUpperBound
    )
    # scrying
    assert (
        original.CalcMrcaUncertaintyAbsExact
        == stripped.CalcMrcaUncertaintyAbsExact
    )
    assert (
        original.CalcMrcaUncertaintyRelExact
        == stripped.CalcMrcaUncertaintyRelExact
    )
    assert (
        original.CalcNumStrataRetainedExact
        == stripped.CalcNumStrataRetainedExact
    )
    assert original.IterRetainedRanks == stripped.IterRetainedRanks
    # enactment
    assert original.GenDropRanks == stripped.GenDropRanks

    # test chaining
    assert (
        pseudostochastic_algo.Policy(
            hash_salt,
        ).WithoutCalcRankAtColumnIndex()
        == stripped
    )


def test_repr():
    hash_salt = 1
    policy = pseudostochastic_algo.Policy(hash_salt)
    assert str(hash_salt) in repr(policy)
    assert policy.GetSpec().GetAlgoIdentifier() in repr(policy)


def test_str():
    hash_salt = 1
    policy = pseudostochastic_algo.Policy(hash_salt)
    assert str(hash_salt) in str(policy)
    assert policy.GetSpec().GetAlgoTitle() in str(policy)



import pandas as pd

from ._alifestd_mark_max_descendant_origin_time_asexual import (
    alifestd_mark_max_descendant_origin_time_asexual,
)


def alifestd_mark_clade_duration_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Add column `clade_duration`, containing the difference between each the
    `origin_time` of each node and the maximum `origin_time` of its descendants.

    Leaf nodes will have duration 0.

    Dataframe reindexing (e.g., df.index) may be applied.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    if "origin_time" not in phylogeny_df.columns:
        raise ValueError("phylogeny_df must contain `origin_time` column")

    if "max_descendant_origin_time" not in phylogeny_df.columns:
        phylogeny_df = alifestd_mark_max_descendant_origin_time_asexual(
            phylogeny_df, mutate=True
        )

    phylogeny_df["clade_duration"] = (
        phylogeny_df["max_descendant_origin_time"].values
        - phylogeny_df["origin_time"].values
    )

    return phylogeny_df



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_delete_unifurcating_roots_asexual,
    alifestd_try_add_ancestor_list_col,
    alifestd_validate,
)

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple1(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[]"],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(df, mutate=mutate)
    assert alifestd_validate(alifestd_try_add_ancestor_list_col(result))
    assert len(result) == 1

    if not mutate:
        assert original_df.equals(df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple2(
    mutate: bool,
):
    df = pd.DataFrame(
        {
            "id": [0, 8],
            "ancestor_id": [0, 0],
            "ancestor_list": ["[none]", "[0]"],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(
        df, root_ancestor_token=""
    )
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [8],
            "ancestor_id": [8],
            "ancestor_list": ["[]"],
        }
    )
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True), expected
    )

    if not mutate:
        assert original_df.equals(df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple3(
    mutate: bool,
):
    df = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_id": [0, 0],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(
        df, root_ancestor_token=""
    )
    assert alifestd_validate(alifestd_try_add_ancestor_list_col(result))

    expected = pd.DataFrame(
        {
            "id": [1],
            "ancestor_id": [1],
        }
    )
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True), expected
    )

    if not mutate:
        assert original_df.equals(df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple4(
    mutate: bool,
):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 0],
            "ancestor_list": ["[none]", "[0]", "[0]"],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(df)
    assert alifestd_validate(result)

    expected = original_df
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True), expected
    )

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple5(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 4, 3, 5],
            "ancestor_id": [0, 0, 1, 2, 0, 5],
            "ancestor_list": ["[none]", "[0]", "[1]", "[2]", "[0]", "[none]"],
            "asdf": ["a", "b", "c", "d", "e", "f"],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(df)
    assert alifestd_validate(result)

    expected = original_df
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True), expected
    )

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple6(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_id": [0, 0, 1, 2, 3, 4, 5],
            "origin_time": [0, 0, 1, 2, 3, 4, 5],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(df)
    assert alifestd_validate(alifestd_try_add_ancestor_list_col(result))

    expected = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5, 6],
            "ancestor_id": [1, 1, 2, 3, 4, 5],
            "origin_time": [0, 1, 2, 3, 4, 5],
        }
    )
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True), expected
    )

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple7(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_id": [0, 0, 1, 2, 3, 4, 0],
            "asdf": [True, True, True, False, True, True, False],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(df)
    assert alifestd_validate(alifestd_try_add_ancestor_list_col(result))

    expected = original_df
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True),
        expected.reset_index(drop=True),
    )

    if not mutate:
        assert df.equals(original_df)


@pytest.mark.parametrize("mutate", [True, False])
def test_alifestd_delete_unifurcating_roots_asexual_simple8(mutate: bool):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 4, 3, 5],
            "ancestor_id": [0, 0, 1, 2, 2, 5],
            "ancestor_list": ["[none]", "[0]", "[1]", "[2]", "[2]", "[none]"],
            "asdf": ["a", "b", "c", "d", "e", "f"],
        }
    )
    original_df = df.copy()
    result = alifestd_delete_unifurcating_roots_asexual(df)
    assert alifestd_validate(result)

    expected = pd.DataFrame(
        {
            "id": [1, 2, 4, 3, 5],
            "ancestor_id": [1, 1, 2, 2, 5],
            "ancestor_list": ["[none]", "[1]", "[2]", "[2]", "[none]"],
            "asdf": ["b", "c", "d", "e", "f"],
        }
    )
    pd.testing.assert_frame_equal(
        result[expected.columns].reset_index(drop=True), expected
    )

    if not mutate:
        assert df.equals(original_df)



import polars as pl
from tqdm import tqdm

from hstrat._auxiliary_lib import (
    alifestd_is_chronologically_ordered,
    alifestd_try_add_ancestor_list_col,
    alifestd_validate,
)
from hstrat.phylogenetic_inference.tree._impl._build_tree_searchtable_cpp_impl_stub import (
    Records,
    build_tree_searchtable_cpp_from_exploded,
    collapse_unifurcations,
    copy_records_to_dict,
    placeholder_value,
)


def test_collapse_unifurcations_singleton():
    records = Records(1)
    records = collapse_unifurcations(records, dropped_only=False)
    assert len(records) == 1


def test_collapse_all_unifurcations_linear_tree():
    # 0 <- 1 <- 2 <- 3 <- 4 <- 5 <- 6 <- 7
    records = Records(8)
    records.addRecord(1, 1, 0, 0, 2, 1, 1, 1, 1)
    records.addRecord(2, 2, 1, 1, 3, 2, 2, 2, 2)
    records.addRecord(3, 3, 2, 2, 4, 3, 3, 3, 3)
    records.addRecord(4, 4, 3, 3, 5, 4, 4, 4, 4)
    records.addRecord(5, 5, 4, 4, 6, 5, 5, 5, 5)
    records.addRecord(6, 6, 5, 5, 7, 6, 6, 6, 6)
    records.addRecord(7, 7, 6, 6, 7, 7, 7, 7, 7)

    records = collapse_unifurcations(records, dropped_only=False)
    assert len(records) == 2

    result = copy_records_to_dict(records)

    # 0 <- 7(1)
    expected = {
        "dstream_data_id": [placeholder_value, 7],
        "id": [0, 1],
        "ancestor_id": [0, 0],
        "search_ancestor_id": [placeholder_value] * 2,
        "search_first_child_id": [placeholder_value] * 2,
        "search_prev_sibling_id": [placeholder_value] * 2,
        "search_next_sibling_id": [placeholder_value] * 2,
        "rank": [0, 7],
        "differentia": [0, 7],
    }

    for key in expected:
        expected_value = expected[key]
        assert result[key] == expected_value, key


def test_collapse_all_unifurcations_branched_tree():
    # 0 <- 1 <- 3
    #  \ <- 2 <- 4 <- 5 <- 6
    #                  \ <- 7
    records = Records(8)
    records.addRecord(1, 1, 0, 0, 3, 1, 2, 1, 1)
    records.addRecord(2, 2, 0, 0, 4, 1, 2, 1, 2)
    records.addRecord(3, 3, 1, 1, 3, 3, 3, 2, 1)
    records.addRecord(4, 4, 2, 2, 5, 4, 4, 2, 1)
    records.addRecord(5, 5, 4, 4, 5, 5, 5, 3, 0)
    records.addRecord(6, 6, 5, 5, 6, 6, 7, 4, 1)
    records.addRecord(7, 7, 5, 5, 7, 6, 7, 4, 2)
    records = collapse_unifurcations(records, dropped_only=False)
    assert len(records) == 5

    result = copy_records_to_dict(records)

    # 0 <- 3(1)
    #  \ <- 5(2) <- 6(3)
    #         \  <- 7(4)
    expected = {
        "dstream_data_id": [placeholder_value, 3, 5, 6, 7],
        "id": [0, 1, 2, 3, 4],
        "ancestor_id": [0, 0, 0, 2, 2],
        "search_ancestor_id": [placeholder_value] * 5,
        "search_first_child_id": [placeholder_value] * 5,
        "search_prev_sibling_id": [placeholder_value] * 5,
        "search_next_sibling_id": [placeholder_value] * 5,
        "rank": [0, 2, 3, 4, 4],
        "differentia": [0, 1, 0, 1, 2],
    }

    for key, value in result.items():
        expected_value = expected[key]
        assert value == expected_value, key


def test_regression_original():
    exploded = pl.DataFrame(
        {
            "dstream_data_id": pl.Series(
                "dstream_data_id",
                [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1],
                dtype=pl.UInt64,
            ),
            "dstream_T": pl.Series(
                "dstream_T",
                [8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11],
                dtype=pl.UInt64,
            ),
            "dstream_value_bitwidth": pl.Series(
                "dstream_value_bitwidth",
                [8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8],
                dtype=pl.UInt32,
            ),
            "dstream_value": pl.Series(
                "dstream_value",
                [0, 1, 3, 7, 2, 5, 4, 6, 0, 1, 3, 7, 2, 5, 9, 6],
                dtype=pl.UInt64,
            ),
            "dstream_Tbar": pl.Series(
                "dstream_Tbar",
                [0, 1, 3, 7, 2, 5, 4, 6, 0, 1, 3, 7, 2, 5, 9, 6],
                dtype=pl.UInt64,
            ),
        },
    ).select(
        pl.col(
            "dstream_data_id",
            "dstream_T",
            "dstream_Tbar",
            "dstream_value",
        )
        .sort_by("dstream_Tbar")
        .over(partition_by="dstream_data_id"),
    )
    res = build_tree_searchtable_cpp_from_exploded(
        exploded["dstream_data_id"].to_numpy(),
        exploded["dstream_T"].to_numpy(),
        exploded["dstream_Tbar"].to_numpy(),
        exploded["dstream_value"].to_numpy(),
        tqdm,
    )
    res = pl.DataFrame(res)
    assert alifestd_validate(
        alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert alifestd_is_chronologically_ordered(
        alifestd_try_add_ancestor_list_col(
            res.with_columns(origin_time=pl.col("rank")).to_pandas(),
        ),
    )


def test_regression_distilled():
    exploded = pl.DataFrame(
        {
            "dstream_data_id": pl.Series(
                "dstream_data_id",
                [0, 0, 1],
                dtype=pl.UInt64,
            ),
            "dstream_T": pl.Series(
                "dstream_T",
                [8, 8, 11],
                dtype=pl.UInt64,
            ),
            "dstream_value_bitwidth": pl.Series(
                "dstream_value_bitwidth",
                [
                    8,
                    8,
                    8,
                ],
                dtype=pl.UInt32,
            ),
            "dstream_value": pl.Series(
                "dstream_value",
                [7, 6, 6],
                dtype=pl.UInt64,
            ),
            "dstream_Tbar": pl.Series(
                "dstream_Tbar",
                [7, 6, 6],
                dtype=pl.UInt64,
            ),
        },
    ).select(
        pl.col(
            "dstream_data_id",
            "dstream_T",
            "dstream_Tbar",
            "dstream_value",
        )
        .sort_by("dstream_Tbar")
        .over(partition_by="dstream_data_id"),
    )
    res = build_tree_searchtable_cpp_from_exploded(
        exploded["dstream_data_id"].to_numpy(),
        exploded["dstream_T"].to_numpy(),
        exploded["dstream_Tbar"].to_numpy(),
        exploded["dstream_value"].to_numpy(),
        tqdm,
    )
    res = pl.DataFrame(res)
    assert alifestd_validate(
        alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )
    assert alifestd_is_chronologically_ordered(
        alifestd_try_add_ancestor_list_col(
            res.with_columns(origin_time=pl.col("rank")).to_pandas(),
        ),
    )



import os

import polars as pl

from hstrat._auxiliary_lib import alifestd_count_leaf_nodes
from hstrat.dataframe import surface_test_drive

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_smoke():
    df = pl.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    df = df.with_columns(foo=pl.lit("bar"))
    pop = surface_test_drive(
        df,
        dstream_algo="dstream.steady_algo",
        dstream_S=64,
        stratum_differentia_bit_width=1,
    )
    assert len(pop) == alifestd_count_leaf_nodes(df)
    assert "origin_time" in pop.columns
    assert "data_hex" in pop.columns
    assert (pop["dstream_storage_bitwidth"] == 64).all()
    assert "dstream_storage_bitoffset" in pop.columns
    assert "dstream_T_bitwidth" in pop.columns
    assert "dstream_T_bitoffset" in pop.columns
    assert "dstream_S" in pop.columns
    assert "dstream_algo" in pop.columns
    assert "downstream_version" in pop.columns
    assert "foo" in pop.columns
    assert "td_source_id" in pop.columns
    assert set(zip(pop["td_source_id"], pop["origin_time"])) <= set(
        zip(df["id"], df["origin_time"]),
    )



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_is_asexual,
    alifestd_is_working_format_asexual,
    alifestd_to_working_format,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_is_working_format(phylogeny_df):
    phylogeny_df_ = phylogeny_df.copy()
    working_df = alifestd_to_working_format(phylogeny_df)

    if alifestd_is_working_format_asexual(phylogeny_df):
        assert working_df.equals(phylogeny_df)
    elif alifestd_is_asexual(phylogeny_df):
        assert not working_df.equals(phylogeny_df)

    if alifestd_is_asexual(phylogeny_df):
        assert alifestd_is_working_format_asexual(working_df)
    else:
        assert not alifestd_is_working_format_asexual(working_df)

    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)
